from __future__ import annotations

from typing import Any

import myke
import yapx

from ..query import _exec_docker_json_query, _jq_query, _print_results


def service_logs(name: str):
    myke.run(["docker", "service", "logs", "-f", "--tail", "100", name])


def service_restart(names: list[str] = yapx.arg(pos=True)):
    for x in names:
        myke.run(
            [
                "docker",
                "service",
                "update",
                "--force",
                "--with-registry-auth",
                "--update-parallelism",
                "0",
                x,
            ],
        )


def service_scale(
    names: list[str] = yapx.arg(pos=True),
    replicas: int = yapx.arg(),
    no_wait: bool = yapx.arg(False, flags=["-W", "--no-wait"]),
):
    myke.run(
        [
            "docker",
            "service",
            "scale",
        ]
        + (["--detach"] if no_wait else [])
        + [f"{x}={replicas}" for x in names],
    )


def service_ls(
    stack: str | None = yapx.arg(None, pos=True),
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    verbose: bool = False,
    _echo: bool = True,
):
    if not verbose and not select:
        select = ["Name", "Mode", "Replicas"]

    results = _exec_docker_json_query(
        ["service", "ls"]
        + (["--filter", "label=com.docker.stack.namespace=" + stack] if stack else []),
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def service_ps(
    names: list[str] | None = yapx.arg(None, pos=True, exclusive=True),
    stacks: list[str] | None = yapx.arg(None, exclusive=True),
    all_states: bool = False,
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    verbose: bool = False,
    _echo: bool = True,
):
    if not verbose and not select:
        select = ["Name", "DesiredState", "CurrentState", "Node", "Error"]

    if not names:
        if stacks:
            names = [
                y
                for x in stacks
                for y in service_ls(
                    stack=x,
                    jq=[],
                    no_table=True,
                    extract="Name",
                    _echo=False,
                )
            ]
        else:
            names = service_ls(jq=[], no_table=True, extract="Name", _echo=False)
    elif isinstance(names, str):
        names = [names]

    assert names
    names.sort()

    state_filters: list[str] = (
        []
        if all_states
        else [
            "--filter",
            "desired-state=running",
            "--filter",
            "desired-state=ready",
            "--filter",
            "desired-state=accepted",
        ]
    )

    results = _exec_docker_json_query(
        ["service", "ps", "--no-trunc"] + state_filters + names,
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def _service_node_containers(name: str) -> list[tuple[str, str]]:
    return [
        (node, container["ID"])
        for node in service_ps(
            [name],
            stacks=None,
            jq=[],
            no_table=True,
            extract="Node",
            _echo=False,
        )
        for container in _exec_docker_json_query(
            [
                "ps",
                "--filter",
                "label=com.docker.swarm.service.name=" + name,
            ],
            env_update={"DOCKER_CONTEXT": node},
        )
    ]


def service_health(
    name: str = yapx.arg(pos=True),
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
):
    results: list[dict[str, Any]] = [
        c_info
        for node, c_id in _service_node_containers(name)
        for c_info in _exec_docker_json_query(
            [
                "inspect",
                c_id,
            ],
            env_update={"DOCKER_CONTEXT": node},
        )
    ]

    results = _jq_query(
        records=results,
        expr=[r".[] | {ID: .Id, Name: .Name, RestartCount: .RestartCount} + .State"]
        + jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def service_ssh(name: str = yapx.arg(pos=True), shell: str = "sh"):
    for node, c_id in _service_node_containers(name):
        print(f"\n*** {node}: {c_id}\n")
        myke.run(
            ["docker", "exec", "-it", c_id, shell],
            env_update={"DOCKER_CONTEXT": node},
        )



from __future__ import annotations

import os
from contextlib import contextmanager

import myke


def current_commit_sha() -> str:
    git_commit_sha, _, _ = myke.run(
        ["git", "rev-parse", "--short", "HEAD"],
        capture_output=True,
        echo=False,
    )
    assert git_commit_sha
    return git_commit_sha.strip()


def current_project_name(from_remote: str = "origin") -> str:
    stdout, _, _ = myke.run(
        ["git", "remote", "get-url", from_remote],
        capture_output=True,
        echo=False,
        check=True,
    )

    assert stdout

    proj_grp, proj_name = stdout.strip().split("/")[-2:]
    if proj_name.endswith(".git"):
        proj_name = os.path.splitext(proj_name)[0]
    proj_name = f"{proj_grp}/{proj_name}"

    return proj_name


@contextmanager
def set_cwd(path: str):
    path_og = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(path_og)


@contextmanager
def set_env(**kwargs):
    env_og = os.environ.copy()
    try:
        for k, v in kwargs.items():
            if v is None:
                os.environ.pop(k)
            else:
                os.environ[k] = v

        yield

    finally:
        os.environ.clear()
        os.environ.update(env_og)



from __future__ import annotations

import os
import sys

import myke
import yapx

from .__version__ import __version__
from .commands.ci import ci_secrets, ci_template, ci_url
from .commands.cluster import (
    cluster_create,
    cluster_edit,
    cluster_exec,
    cluster_groups,
    cluster_ls,
    cluster_ping,
    cluster_playbook,
    cluster_query,
    cluster_setup,
    cluster_ssh,
)
from .commands.context import context_current, context_default, context_ls
from .commands.registry import (
    registry_image_tags,
    registry_images,
    registry_import,
    registry_login,
    registry_ls,
)
from .commands.service import (
    service_health,
    service_logs,
    service_ls,
    service_ps,
    service_restart,
    service_scale,
    service_ssh,
)
from .commands.stack import (
    stack_build,
    stack_config,
    stack_create,
    stack_down,
    stack_ls,
    stack_ps,
    stack_secrets,
    stack_up,
    stack_update,
    stack_wait,
)
from .commands.swarm import (
    swarm_exec,
    swarm_labels,
    swarm_networks,
    swarm_nodes,
    swarm_ping,
    swarm_prune,
    swarm_query,
    swarm_reboot,
    swarm_ssh,
    swarm_stats,
)
from .query import _exec_json_query, _print_results
from .utils import set_env

ANSIBLE_COLLECTION_DIR: str = os.path.join(os.path.dirname(__file__), "collections")
ANSIBLE_REQUIREMENTS_FILE: str = os.path.join(
    ANSIBLE_COLLECTION_DIR,
    "requirements.yml",
)


def setup(
    cluster: str | None = yapx.arg(None, env="DOCKER_CONTEXT"),
    inventory_dir: str
    | None = yapx.arg(
        None,
        env="HOSTBUTTER_INVENTORY_DIR",
    ),
    docker_config: str | None = yapx.arg(None, env="DOCKER_CONFIG"),
):
    if not docker_config:
        docker_config = os.path.join(os.path.expanduser("~"), ".docker")

    docker_config_file: str = os.path.join(docker_config, "config.json")

    with set_env(
        **{
            "DOCKER_CONFIG": docker_config,
            "DOCKER_CONFIG_FILE": docker_config_file,
        },
    ):
        if not cluster:
            cluster = context_current(from_env=True, _echo=False)
            assert cluster

        if not inventory_dir:
            inventory_dir = os.path.join(os.path.expanduser("~"), ".hostbutter")

        ansible_inv_file: str = os.path.join(inventory_dir, cluster, "inventory.yml")

        with set_env(
            **{
                "DOCKER_CONTEXT": cluster,
                "ANSIBLE_INVENTORY": ansible_inv_file,
            },
        ):
            yield


def install_playbooks():
    myke.run(
        ["ansible-galaxy", "role", "install", "-r", ANSIBLE_REQUIREMENTS_FILE],
    )
    myke.run(
        ["ansible-galaxy", "collection", "install", "-r", ANSIBLE_REQUIREMENTS_FILE],
    )
    myke.run(
        [
            "ansible-galaxy",
            "collection",
            "install",
            "--force",
            os.path.join(
                ANSIBLE_COLLECTION_DIR,
                "ansible_collections",
                "fresh2dev",
                "hostbutter",
            ),
        ],
    )


def sh_exec(command: str):
    myke.run(command)


def sh_json(
    command: str,
    jq: list[str] | None = None,
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
):
    results = _exec_json_query(
        cmd=command,
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


# def registry_ls():
#     myke.read.json(os.environ["DOCKER_CONFIG_FILE"])

# def tmp_reg_creds():
#     myke.echo.pretty(myke.read.json(os.environ["DOCKER_CONFIG_FILE"]))


# TODO: implement
# def stack-catalog:
# def ci-(url|secrets|template|deploy):
# organize/segment code into private/public
# THEN, retire .hostbutter.bashrc


def main() -> None:
    args: list[str] = sys.argv[1:]
    yapx.run(
        setup,
        cluster_playbook,
        cluster_setup,
        cluster_groups,
        cluster_query,
        cluster_create,
        cluster_edit,
        cluster_ls,
        cluster_ping,
        cluster_exec,
        cluster_ssh,
        context_ls,
        context_current,
        context_default,
        swarm_stats,
        swarm_ssh,
        swarm_nodes,
        swarm_labels,
        swarm_ping,
        swarm_exec,
        swarm_prune,
        swarm_reboot,
        swarm_networks,
        stack_ls,
        stack_ps,
        stack_config,
        stack_build,
        stack_up,
        stack_down,
        stack_wait,
        stack_secrets,
        stack_update,
        stack_create,
        service_logs,
        service_ls,
        service_ps,
        service_restart,
        service_ssh,
        service_scale,
        service_health,
        registry_login,
        registry_ls,
        registry_images,
        registry_import,
        registry_image_tags,
        ci_url,
        ci_template,
        ci_secrets,
        docker=swarm_query,
        sh=sh_exec,
        sh_json=sh_json,
        version=lambda: print(__version__),
        install=install_playbooks,
        _args=args,
        _print_help="--help-full" in args,
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import os
from fnmatch import fnmatch
from getpass import getpass
from typing import Any

import myke
import requests
import yapx

from ..constants import CI_IMAGE_DEPENDENCIES, IMAGE_DEPENDENCIES
from ..exceptions import NoKnownRegistryError, NoTagsFoundError, TooManyTagsError
from ..query import _exec_json_query, _jq_query, _print_results
from .context import context_current
from .swarm import swarm_ssh

REGISTRY_NAME_MAP: dict[str, str] = {"docker.io": "https://index.docker.io/v1/"}


def _skopeo_cmd(skopeo_args: str | list[str], src_registry: str | None = None) -> str:
    ctx: str = context_current(from_env=True, _echo=False)

    if not src_registry:
        src_registry = _get_default_registry(ctx)

    cmd: str = (
        r"""
docker run --rm -i \
    -v /etc/ssl/certs/ca-certificates.crt:/etc/ssl/certs/ca-certificates.crt:ro \
    -v "${DOCKER_CONFIG:-$HOME/.docker}/config.json:/.docker/config.json:ro" \
    -e REGISTRY_AUTH_FILE=/.docker/config.json
""".strip()
        + f" {src_registry}/skopeo/stable:v1.9.2"
        + " "
    )

    if not isinstance(skopeo_args, str):
        skopeo_args = " ".join(skopeo_args)

    ctx_ssh: str = swarm_ssh(name=ctx, dry_run=True, _echo=False)
    if ctx_ssh.startswith("ssh://"):
        cmd = f"ssh {ctx_ssh} '{cmd} {skopeo_args}'"
    else:
        cmd += skopeo_args

    return cmd


def _get_default_registry(context: str | None = None) -> str:
    reg: str | None = os.getenv("HB_IMAGE_REGISTRY")

    if not reg:
        if not context:
            context = context_current(from_env=True, _echo=False)

        if "." not in context:
            raise NoKnownRegistryError(
                "Unable to infer registry URL within current context: " + context,
            )

        reg = "registry." + context

    return reg


def _split_image_parts(
    image_name: str,
    default_registry: str | None = None,
    default_tag: str | None = None,
) -> tuple[str, str, str | None]:
    if not default_registry:
        default_registry = "docker.io"

    image_parts: list[str] = image_name.split(":", maxsplit=1)

    image_name = image_parts[0]

    image_tag: str | None = image_parts[1] if len(image_parts) > 1 else default_tag

    image_parts = image_name.split("/", maxsplit=1)

    if len(image_parts) == 1:
        image_registry = default_registry
        image_name = image_parts[0]
    elif "." not in image_parts[0]:
        image_registry = default_registry
    else:
        image_registry = image_parts[0]
        image_name = image_parts[1]

    return image_registry, image_name, image_tag


def registry_ls(
    raw: bool = False,
    _echo: bool = True,
) -> list[dict[str, str]] | list[str]:
    ctx: str = context_current(from_env=True, _echo=False)
    ctx_ssh: str = swarm_ssh(name=ctx, dry_run=True, _echo=False)

    is_ssh_context: bool = ctx_ssh.startswith("ssh://")
    cmd: list[str] = ["ssh", ctx_ssh] if is_ssh_context else []
    cmd.append("cat ${DOCKER_CONFIG:-~/.docker}/config.json")

    stdout, _, _ = myke.run(
        cmd,
        capture_output=True,
        echo=False,
        shell=(not is_ssh_context),
    )
    assert stdout

    reg_info: dict[str, dict[str, str]] | list[dict[str, str]] | list[str] = json.loads(
        stdout,
    )

    if not raw:
        reg_info = list(reg_info["auths"].keys())

    if _echo:
        _print_results(reg_info, no_table=True, pretty=False, raw=False)

    return reg_info


def _get_registry_creds(name: str) -> str:
    results: dict[str, Any] = registry_ls(raw=True, _echo=False)

    matching: list[str] = [
        v["auth"]
        for k, v in results["auths"].items()
        if k == REGISTRY_NAME_MAP.get(name, name)
    ]

    if not matching:
        print("Unrecognized registry. Select one of:")
        registry_ls(raw=False, _echo=True)
        raise ValueError(name)

    return matching[0]


def registry_login(
    url: str | None = yapx.arg(None, pos=True, env="HB_IMAGE_REGISTRY"),
    username: str = yapx.arg(None, env="HB_IMAGE_REGISTRY_USER"),
    password: str | None = yapx.arg(None, env="HB_IMAGE_REGISTRY_PASS"),
):
    ctx: str = context_current(from_env=True, _echo=False)
    ctx_ssh: str = swarm_ssh(name=ctx, dry_run=True, _echo=False)

    if not url:
        url = _get_default_registry(ctx)

    if not username:
        username = input("Username: ")
        assert username

    if not password:
        password = getpass()
        assert password

    is_ssh_context: bool = ctx_ssh.startswith("ssh://")
    cmd: list[str] = ["ssh", ctx_ssh] if is_ssh_context else []
    cmd.append(f'docker login --username "{username}" --password-stdin {url}')

    myke.run(cmd, input=password.encode(), shell=(not is_ssh_context))


def registry_images(
    registry: str | None = yapx.arg(None, pos=True),
    img_filter: str | None = yapx.arg(None, flags=["-f", "--filter"]),
    v1_api: bool = yapx.arg(False, flags="--v1-api"),
    limit: int = yapx.arg(1000, flags=["-n", "--limit"]),
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    verbose: bool = False,
    _echo: bool = True,
) -> list[str]:
    if not registry:
        registry = _get_default_registry()

    reg_url: str = REGISTRY_NAME_MAP.get(registry, registry)

    if not reg_url.startswith("https://"):
        reg_url = f"https://{reg_url}/" + ("v1" if v1_api else "v2")

    reg_url = reg_url.rstrip("/")

    is_v1_api: bool = "v1" in reg_url
    if is_v1_api:
        if not img_filter:
            raise ValueError("Image filter is required for Registry v1 API.")
        reg_url = reg_url + f"/search?q={img_filter.replace('*', '')}&n=100"
    else:
        reg_url = reg_url + "/_catalog"

    imgs: list[dict[str, str | int | bool]] = []

    limit_reached: bool = False

    page: int = 0
    num_pages: int = 0
    while not limit_reached and (page == 0 or (is_v1_api and page < num_pages)):
        resp: requests.Response = requests.get(
            reg_url + (f"&page={page+1}" if page > 0 else ""),
            headers={"Authorization": f"Basic {_get_registry_creds(name=registry)}"},
            verify=os.getenv(
                "REQUESTS_CA_BUNDLE",
                "/etc/ssl/certs/ca-certificates.crt",
            ),
            timeout=300,
        )

        resp.raise_for_status()

        resp_data: dict[str, Any] = resp.json()

        if is_v1_api:
            imgs.extend(resp_data["results"])

            page = resp_data["page"]
            num_pages = resp_data["num_pages"]
        else:
            imgs.extend({"name": x} for x in resp_data["repositories"])
            page += 1

        if img_filter:
            imgs = [x for x in imgs if fnmatch(x["name"], img_filter)]

        if limit > 0 and len(imgs) >= limit:
            limit_reached = True

    for x in imgs:
        x["name"] = f"{registry}/{x['name']}"

    if is_v1_api and not verbose and not select:
        select = [
            "name",
            "pull_count",
            "star_count",
        ]

    if not sort:
        if is_v1_api:
            sort = ["pull_count", "star_count"]
        else:
            sort = ["name"]
            sort_reverse = True

    results = _jq_query(
        records=imgs,
        expr=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if limit_reached:
        results = results[:limit]

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

        if limit_reached:
            print(
                (
                    '...\n\n*** Result limit reached. Adjust filter ("--filter") and/or'
                    ' limit ("--limit").\n'
                ),
            )

    return results


def registry_image_tags(
    image_name: str = yapx.arg(pos=True),
    image_registry: str | None = yapx.arg(None, flags=["-r", "--registry"]),
    tag_filter: str | None = yapx.arg(None, flags=["-f", "--filter"]),
    remove: bool = yapx.arg(False, flags=["--rm", "--remove"]),
    _echo: bool = True,
    _skopeo_image_registry: str | None = None,
) -> list[str]:
    new_img_registry, image_name, image_tag = _split_image_parts(
        image_name=image_name,
        default_registry=image_registry,
        default_tag=None,
    )

    if image_registry:
        new_img_registry = image_registry

    if not tag_filter and image_tag:
        tag_filter = image_tag

    image_name_fq: str = f"{new_img_registry}/{image_name}"

    results: list[str] = [
        y
        for x in _exec_json_query(
            _skopeo_cmd(
                "list-tags --retry-times 2 docker://" + image_name_fq,
                src_registry=_skopeo_image_registry,
            ),
        )
        for y in x["Tags"]
    ]

    if tag_filter:
        results = [x for x in results if fnmatch(x, tag_filter)]

    results = [f"{image_name_fq}:{x}" for x in results]

    if remove:
        for x in results:
            if _echo:
                print("Deleting image tag: " + x)
            myke.run(
                _skopeo_cmd(
                    "delete docker://" + x,
                    src_registry=_skopeo_image_registry,
                ),
            )
    elif _echo:
        myke.echo.lines(results)

    return results


def registry_import(
    image_names: list[str] | None = yapx.arg(None, pos=True, exclusive=True),
    import_file: str | None = yapx.arg(None, exclusive=True, flags=["-f", "--file"]),
    dependencies: bool | None = yapx.arg(None, exclusive=True),
    ci_dependencies: bool | None = yapx.arg(None, exclusive=True),
    new_registry: str | None = yapx.arg(None, flags=["-r", "--registry"]),
    new_owner: str | None = yapx.arg(None, flags=["-o", "--owner"]),
    all_tags: bool = yapx.arg(False, flags=["-a", "--all-tags"]),
    src_no_creds: bool = False,
    multi_arch: bool = False,
    force: bool = False,
    dry_run: bool = False,
    skopeo_image_registry: str | None = yapx.arg(None, flags=["--skopeo-registry"]),
):
    import_images: list[str] = (
        image_names
        if image_names
        else [
            y
            for x in myke.read.lines(import_file)
            for y in [x.strip()]
            if not y.startswith("#")
        ]
        if import_file
        else IMAGE_DEPENDENCIES
        if dependencies
        else CI_IMAGE_DEPENDENCIES
        if ci_dependencies
        else []
    )

    if not new_registry:
        new_registry = _get_default_registry()

    for this_image in import_images:
        src_registry, img_name, img_tag = _split_image_parts(image_name=this_image)

        src_images: list[str] = (
            [f"{src_registry}/{img_name}:{img_tag}"]
            if img_tag and "*" not in img_tag
            else registry_image_tags(
                img_name,
                image_registry=src_registry,
                tag_filter=img_tag,
                remove=False,
                _echo=False,
                _skopeo_image_registry=skopeo_image_registry,
            )
        )

        if not src_images:
            raise NoTagsFoundError(
                f"No matching tags found for image: {src_registry}/{img_name}",
            )

        if not img_tag and len(src_images) > 1 and not all_tags:
            raise TooManyTagsError(
                (
                    "No tag given and multiple tags found."
                    " Specify a tag, or use '--all-tags' to import all tags."
                ),
            )

        if new_owner:
            img_name_parts: list[str] = img_name.split("/", maxsplit=1)
            img_name = f"{new_owner}/{img_name_parts[len(img_name_parts)-1]}"

        if not skopeo_image_registry and (dependencies or ci_dependencies):
            skopeo_image_registry = "quay.io"

        tgt_images: list[str]

        try:
            tgt_images = registry_image_tags(
                img_name,
                image_registry=new_registry,
                tag_filter=None,
                remove=False,
                _echo=False,
                _skopeo_image_registry=skopeo_image_registry,
            )
        except myke.exceptions.CalledProcessError as e:
            print(e.stdout)
            print(e.stderr)
            print(e.returncode)
            tgt_images = []

        for src_img in src_images:
            _, src_tag = src_img.split(":", maxsplit=1)
            tgt_img: str = f"{new_registry}/{img_name}:{src_tag}"

            if not force and tgt_img in tgt_images:
                print("exists: " + tgt_img)
            else:
                print("import: " + tgt_img)
                if not dry_run:
                    extra_args: list[str] = ["--retry-times", "2"]
                    if src_no_creds:
                        extra_args.append("--src-no-creds")
                    if multi_arch:
                        extra_args.extend(["--multi-arch", "all"])

                    myke.run(
                        _skopeo_cmd(
                            (
                                "copy"
                                f" {' '.join(extra_args)} docker://{src_img} docker://{tgt_img}"
                            ),
                            src_registry=skopeo_image_registry,
                        ),
                    )



from __future__ import annotations

import os
import sys
from typing import Any

import myke
import yapx

from ..query import (
    _exec_ansible_json_query,
    _exec_ansibleinv_json_query,
    _print_results,
)
from ..utils import set_cwd
from .context import context_current

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


def cluster_playbook(
    name: str,
    limit_to: str = "all",
    password_auth: bool = False,
    _ansible_args: list[str] | None = None,
):
    with set_cwd(os.path.dirname(os.environ["ANSIBLE_INVENTORY"])):
        pb_args: list[str] = []
        if password_auth:
            pb_args.append("-kK")
        if _ansible_args:
            pb_args.extend(_ansible_args)
        pb_args.append(name)

        myke.run(["ansible-playbook", "-l", limit_to] + pb_args)


def cluster_setup(
    scope: None
    | Literal["ssh", "dns", "firewall", "swarm", "contexts", "labels"] = yapx.arg(
        None,
        pos=True,
    ),
    limit_to: str = "all",
    password_auth: bool = False,
):
    ansible_args: list[str] | None = None

    playbook: str = "fresh2dev.hostbutter.apply_swarm"

    if scope:
        if scope == "ssh":
            playbook = "fresh2dev.hostbutter.apply_ssh"
        else:
            ansible_args = ["--tag"]
            if scope == "swarm":
                ansible_args.append("swarm")
            elif scope == "contexts":
                ansible_args.append("swarm-contexts")
            elif scope == "labels":
                ansible_args.append("swarm-labels")
            elif scope == "firewall":
                ansible_args.append("firewall")
            elif scope == "dns":
                ansible_args.extend(
                    ["dns", "--skip-tags", "dnsmasq-install-dependencies"],
                )
            else:
                raise ValueError("Unknown scope: " + scope)

    cluster_playbook(
        playbook,
        limit_to=limit_to,
        password_auth=password_auth,
        _ansible_args=ansible_args,
    )


def cluster_create():
    inventory: str = os.environ["ANSIBLE_INVENTORY"]
    if os.path.exists(inventory):
        raise FileExistsError(inventory)
    os.makedirs(os.path.dirname(inventory))
    with open(inventory, "w+", encoding="utf-8") as f:
        domain: str = os.getenv("HB_DOMAIN", "example.com")
        f.write(
            f"""---
all:
  vars:
    domain: {domain}
    ansible_python_interpreter: python3
    ansible_port: 22
    ansible_ssh_private_key_file: "~/.ssh/id_ed25519.pub"
    root_ca_cert: ""
    timesync_timezone: America/Chicago
    fail2ban_install: true
    fail2ban_maxretry: 5     # allow up to 5 ssh failures,
    fail2ban_findtime: 1440  # in a 24-hour span,
    fail2ban_bantime: 10080  # ban for 7 days.
    dnsmasq_dns_servers:
      - 1.1.1.1
      - 1.0.0.1
    # dnsmasq_wildcard_addresses:
    #   "{domain}": "192.168.69.2"
    dnsmasq_host_records:
      "nas.{domain}": "192.168.69.2"
      "git.{domain}": "192.168.69.2"
      "registry.{domain}": "192.168.69.2"
    docker_daemon_options:
      dns: []
      default-address-pools:
        - base: "10.10.0.0/16"
          size: 24
      log-driver: "json-file"
      log-opts:
        max-size: "10m"
        max-file: "5"
      metrics-addr: '0.0.0.0:9323'
      experimental: true
  children:
    docker_swarm:
      children:
        docker_swarm_manager:
          hosts:
            swarm-mgr-01:
              ansible_host: "1.234.456.78"
              ansible_user: user
              docker_swarm_interface: enp7s0
              swarm_labels:
                - proxy
                - nas
                - frontend
                - backend
                - git-server
                - collector
              nfs_exports:
                - "/mnt/data            *(rw,fsid=0,crossmnt,sync,no_subtree_check,no_auth_nlm,insecure,no_root_squash)"
              firewall_additional_rules: []

        docker_swarm_worker:
          hosts: {{}}

    non_swarm: {{}}
""",
        )
    myke.run([os.getenv("EDITOR", "vim"), os.environ["ANSIBLE_INVENTORY"]])


def cluster_edit():
    myke.run([os.getenv("EDITOR", "vim"), os.environ["ANSIBLE_INVENTORY"]])


def cluster_ls():
    myke.echo.lines(
        [
            x
            for x in os.listdir(
                os.path.dirname(os.path.dirname(os.environ["ANSIBLE_INVENTORY"])),
            )
            if not x.startswith(".")
        ],
    )


def cluster_query(
    view: Literal[
        "all",
        "os",
        "vol",
        "volumes",
        "chassis",
        "hw",
        "hardware",
        "ip",
        "ips",
        "interfaces",
        "inv",
        "inventory",
        "ssh",
    ] = yapx.arg(pos=True),
    limit_to: str = "all",
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
) -> Any:
    results: Any = None

    if view == "ssh":
        results = _exec_ansible_json_query(
            [
                "-m",
                "shell",
                "-a",
                # r'echo "{\"$(hostname)\": \"{{ ansible_user}}@{{ ansible_host }}:{{ ansible_port }}\"}"',
                (
                    r'echo "{\"Host\": \"$(hostname)\", \"SSH\": \"ssh://{{'
                    r' ansible_user}}@{{ ansible_host }}:{{ ansible_port }}\"}"'
                ),
            ],
            limit_to=limit_to,
            jq=[r". [] | {Name: .__hostname__} + . | del (.__hostname__)"] + jq,
            select=select,
            drop=drop,
            extract=extract,
            sort=sort,
            sort_reverse=sort_reverse,
            unique=unique,
        )

    elif view in ("inv", "inventory"):
        jq_base: str = (
            r".[] | ._meta.hostvars | to_entries[] | {hostname: .key} + .value"
            r" | {hostname, ansible_host, ansible_user, docker_swarm_interface,"
            r" dnsmasq_dns_servers, swarm_labels}"
        )
        results = _exec_ansibleinv_json_query(
            jq=[jq_base] + jq,
            select=select,
            drop=drop,
            extract=extract,
            sort=sort,
            sort_reverse=sort_reverse,
            unique=unique,
        )
    else:
        gather_subset: list[str] = ["!all"]

        jq_base: str = r".[] | .ansible_facts"

        if view == "all":
            gather_subset = ["all"]
        elif view == "os":
            gather_subset = ["lsb", "processor"]
            jq_base += (
                r" | {"
                r" HostName: .ansible_hostname"
                r", OS: .ansible_lsb.description"
                r", Kernel: .ansible_kernel"
                r", Arch: .ansible_architecture"
                # r', User: .ansible_user_id'
                # r', UID: .ansible_user_uid'
                r", UptimeDays: ((10 * .ansible_uptime_seconds/86400 | round)/10.0)"
                r", Time: .ansible_date_time.time"
                r" }"
            )
        elif view in ("vol", "volumes"):
            gather_subset = ["mounts"]
            jq_base += (
                r" | {HostName: .ansible_hostname,"
                r" VolumeMounts: .ansible_mounts[] | select(.size_total > 0)}"
                r" | {"
                r" HostName"
                r", Device: .VolumeMounts.device"
                r", Mount_Point: .VolumeMounts.mount"
                r", Volume_Type: .VolumeMounts.fstype"
                r", Size_GB: ((10 * (.VolumeMounts.size_total/1024/1024/1024) |"
                r" round)/10.0)"
                r", Use_Percent: ((10 * (100 * (.VolumeMounts.size_total -"
                r" .VolumeMounts.size_available) / .VolumeMounts.size_total) |"
                r" round)/10.0)"
                r" }"
            )
        elif view == "chassis":
            gather_subset = ["hardware"]
            jq_base += (
                r" | {"
                r" HostName: .ansible_hostname"
                r", SystemVendor: .ansible_system_vendor"
                r", SystemModel: .ansible_product_name"
                r", FormFactor: .ansible_form_factor"
                r", BoardName: .ansible_board_name"
                r", BiosVersion: .ansible_bios_version"
                r", BiosDate: .ansible_bios_date"
                r" }"
            )
        elif view in ("hw", "hardware"):
            gather_subset = ["hardware"]
            jq_base += (
                r" | {"
                r" HostName: .ansible_hostname"
                # r', System_Vendor: .ansible_system_vendor'
                r", SystemModel: .ansible_product_name"
                r", CPU: .ansible_processor[2]"
                r", CPUCores: .ansible_processor_cores"
                r", RAMTotalGB: ((10*.ansible_memtotal_mb/1024 | round)/10.0)"
                r", RAMUsePercent: ((10*100*.ansible_memfree_mb/.ansible_memtotal_mb |"
                r" round)/10.0)"
                r" }"
            )
        elif view in ("ip", "ips", "interfaces"):
            gather_subset = ["network", "dns"]
            jq_base += (
                r" | {ansible_facts: ., main_interfaces: ([.ansible_interfaces[] |"
                r' select(. == "lo" or startswith("veth") or startswith("docker") |'
                r' not)] | map("ansible_"+ .))}'
                r" | {hostname: .ansible_facts.ansible_hostname, domain:"
                r" .ansible_facts.ansible_domain, ifc:"
                r" .ansible_facts[.main_interfaces[]], dns:"
                r" .ansible_facts.ansible_dns.nameservers}"
                r" | select(.ifc.ipv4.address != null)"
                r" | {HostName: .hostname, Domain: .domain, Interface: .ifc.device,"
                r" IPv4: .ifc.ipv4.address, IPv6: [.ifc.ipv6[]?.address], MAC:"
                r" .ifc.macaddress, DNS: .dns}"
            )
        else:
            raise ValueError("Unrecognized query view: " + view)

        results = _exec_ansible_json_query(
            [
                "-m",
                "setup",
                "-a",
                "gather_subset=" + ",".join(gather_subset),
            ],
            limit_to=limit_to,
            jq=[jq_base] + jq,
            select=select,
            drop=drop,
            extract=extract,
            sort=sort,
            sort_reverse=sort_reverse,
            unique=unique,
        )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def cluster_ssh(
    host: list[str] | None = yapx.arg(None, pos=True, exclusive=True),
    ls: bool = yapx.arg(False, exclusive=True, flags=["-l", "--list"]),
):
    if ls or not host:
        cluster_query(view="ssh", jq=[], _echo=True, no_table=False)
        return

    for h in host:
        if h.endswith(context_current(from_env=True, _echo=False)):
            h = h.split(".", maxsplit=1)[0]

        results: list[str] = cluster_query(
            limit_to=h,
            view="ssh",
            jq=[],
            extract="SSH",
            _echo=False,
        )

        if not results:
            print("Unknown host name. Select from:")
            cluster_query(view="ssh", jq=[], _echo=True, no_table=False)
            break

        assert len(results) == 1

        myke.run(["ssh", results[0]])

    return


def cluster_groups():
    myke.run(["ansible-inventory", "--graph"])


def cluster_ping(limit_to: str = "all", skip_verify: bool = False):
    myke.run(
        [
            "ansible",
            limit_to,
            "-m",
            "ping",
        ],
        env_update={"ANSIBLE_HOST_KEY_CHECKING": "False"} if skip_verify else {},
    )


def cluster_exec(command: str, limit_to: str = "all"):
    myke.run(["ansible", limit_to, "-m", "shell", "-a", command])



from __future__ import annotations

import json
from typing import Any

import jq as _jq
import myke

from .constants import SWARM_GROUP_NAME


def _jq_query(
    records: list[dict[str, Any]],
    expr: str | list[str],
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
) -> Any:
    if isinstance(expr, str):
        expr = [expr]

    if sort:
        sort_expr: str = f'. | sort_by(.{", .".join(sort)})'
        if not sort_reverse:
            sort_expr += " | reverse"
        sort_expr += "[]"
        expr.append(sort_expr)

    if drop:
        expr.append(f'.[] | del(.{", .".join(drop)})')

    if select:
        expr.append(f'.[] | {{{", ".join(select)}}}')

    if extract:
        expr.append(f".[] | .{extract}")

    if unique:
        expr.append(". | unique_by(.)[]")

    for x in expr:
        records = _jq.compile(x).input(records).all()

    return records


def _print_results(
    results: Any,
    no_table: bool = False,
    pretty: bool = False,
    raw: bool = False,
) -> None:
    if results is not None:
        if no_table or (
            isinstance(results, list)
            and results
            and (not isinstance(results[0], dict) or len(results[0].keys()) > 6)
        ):
            if raw:
                myke.echo.lines(results)
            elif pretty:
                myke.echo.pretty(results)
            else:
                myke.echo.json(results, indent=2)
        elif raw:
            myke.echo.table(results, tablefmt="tsv")
        elif pretty:
            myke.echo.table(results, tablefmt="rounded_grid")
        else:
            myke.echo.table(results, tablefmt="github")


def _exec_json_query(
    cmd: str | list[str],
    jq: list[str] | None = None,
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    **kwargs: Any,
) -> Any:
    if jq is None:
        jq = []

    stdout, _, _ = myke.run(cmd, capture_output=True, echo=False, **kwargs)

    records: list[dict[str, Any]]

    try:
        parsed = json.loads(stdout)
        if isinstance(parsed, dict):
            records = [parsed]
        elif isinstance(parsed, list):
            records = parsed
        else:
            raise TypeError(
                "Unsupported Json data structure; list of dictionaries expected.",
            )
    except json.JSONDecodeError:
        try:
            records = [json.loads(x) for x in stdout.splitlines()]
        except json.JSONDecodeError:
            try:
                records = [
                    {
                        **({"__hostname__": hostname} if hostname else {}),
                        **json.loads(payload),
                    }
                    for x in stdout.splitlines()
                    for i in [x.index("{")]
                    for hostname in [x[:i].split("|", maxsplit=1)[0].strip()]
                    for payload in x[i:].split("\\n")
                ]
            except json.JSONDecodeError as e:
                print(stdout[:200])
                raise e

    return _jq_query(
        records,
        expr=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )


def _exec_ansible_json_query(
    cmd: list[str],
    limit_to: str = "all",
    jq: list[str] | None = None,
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    **kwargs: Any,
) -> Any:
    return _exec_json_query(
        cmd=["ansible", limit_to, "--one-line"] + cmd,
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
        **kwargs,
    )


def _exec_ansibleinv_json_query(
    jq: list[str] | None = None,
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    **kwargs: Any,
) -> Any:
    return _exec_json_query(
        cmd=["ansible-inventory", "--list"],
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
        **kwargs,
    )


def _exec_ansible_docker_json_query(
    cmd: list[str],
    limit_to: str = SWARM_GROUP_NAME,
    jq: list[str] | None = None,
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    **kwargs: Any,
):
    return _exec_ansible_json_query(
        [
            "-m",
            "command",
            "-a",
            "docker " + " ".join(cmd) + r""" --format {{ '"{{ json . }}"' }}""",
        ],
        limit_to=limit_to,
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
        **kwargs,
    )


def _exec_docker_json_query(
    cmd: list[str],
    jq: list[str] | None = None,
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    **kwargs: Any,
) -> Any:
    return _exec_json_query(
        cmd=["docker", *cmd, "--format", r"{{ json . }}"],
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
        **kwargs,
    )



from __future__ import annotations

SWARM_GROUP_NAME: str = "docker_swarm"

STACK_NAME_VAR: str = "COMPOSE_PROJECT_NAME"

IMAGE_DEPENDENCIES: list[str] = [
    "quay.io/skopeo/stable:v1.9.2",
    "gcr.io/kaniko-project/executor:v1.9.1",
]

CI_IMAGE_DEPENDENCIES: list[str] = IMAGE_DEPENDENCIES + [
    "docker.io/alpine/git:2.36.3",
    "docker.io/docker:20.10.23-dind",
]



from __future__ import annotations

import os
from dataclasses import dataclass
from functools import lru_cache
from getpass import getuser
from time import time

import myke
import packaging.version
import yapx

from ..constants import STACK_NAME_VAR
from ..query import _exec_docker_json_query, _print_results
from ..utils import current_commit_sha, set_cwd, set_env
from .ci import ci_secrets, current_project_name
from .context import context_current
from .registry import _get_default_registry, registry_import
from .service import service_ps
from .swarm import swarm_ssh


def _kaniko_cli(
    src_registry: str | None = None,
    force_remote_build: bool = False,
) -> str:
    ctx: str = context_current(from_env=True, _echo=False)
    ctx_ssh: str = swarm_ssh(name=ctx, dry_run=True, _echo=False)

    if not src_registry:
        src_registry = _get_default_registry(ctx)

    tar_cmd: str = ""

    if force_remote_build or not ctx_ssh.startswith("unix://"):
        tar_args: list[str] = ["--exclude='.[^/]*'"]
        if os.path.exists(".dockerignore"):
            tar_args.append("--exclude-from='.dockerignore'")
        tar_args.extend(["-cvf", "-", "."])
        tar_cmd = f"COPYFILE_DISABLE=true tar {' '.join(tar_args)} | gzip -9 | \\"

    return (
        tar_cmd
        + r"""
        docker run --rm -i \
            -v /etc/ssl/certs/ca-certificates.crt:/etc/ssl/certs/ca-certificates.crt:ro \
            -v "${DOCKER_CONFIG:-$HOME/.docker}/config.json:/.docker/config.json"
        """.strip()
        + (" -v $(pwd):/workspace" if not tar_cmd else "")
        + f" -e DOCKER_CONFIG=/.docker {src_registry}/kaniko-project/executor:v1.9.1"
        + " --context "
        + ("tar://stdin" if tar_cmd else "dir:///workspace")
        + " "
    )


# TODO: refactor so this not called multiple times
# @lru_cache
def _get_stack_name(path: str | None = None, domain: str | None = None) -> str:
    if path and os.path.sep not in path and not os.path.isdir(path):
        return path

    env_file: str = os.path.join(path if path else os.getcwd(), ".env")

    stack_name: str | None = (
        myke.read.envfile(env_file).get(STACK_NAME_VAR).strip().lower()
    )

    if not stack_name:
        raise KeyError(f"{STACK_NAME_VAR} not defined in {env_file}")

    if domain and context_current(from_env=True, _echo=False) != domain:
        # if context is not equal to domain,
        # prepend domain to stack name.
        stack_name = domain.replace(".", "-") + "_" + stack_name

    return stack_name


def stack_wait(
    names: list[str] | None = yapx.arg(None, pos=True),
    timeout: int = 300,
    suppress_errors: bool = False,
) -> None:
    if not names:
        names = [os.getcwd()]

    stack_wait_script: str = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "resources",
        "docker-stack-wait.sh",
    )
    names = [
        _get_stack_name(x) if (os.path.sep in x or os.path.isdir(x)) else x
        for x in names
    ]

    for x in names:
        myke.run(
            [
                "sh",
                stack_wait_script,
                "-t",
                str(timeout),
                "-p",
                "10",
                x,
            ],
            check=(not suppress_errors),
        )


@dataclass
class _StackEnv:
    name: str | None = None
    domain: str | None = None
    image_registry: str | None = None
    image_owner: str | None = None
    image_tag: str | None = None
    nfs_opts: str | None = None

    def __post_init__(self):
        if not self.domain:
            self.domain = os.getenv("HB_DOMAIN", os.environ["DOCKER_CONTEXT"])

        self.name = _get_stack_name(path=self.name, domain=self.domain)

        if not self.image_registry:
            self.image_registry = os.getenv(
                "HB_IMAGE_REGISTRY",
                "registry." + self.domain,
            )

        if not self.image_owner:
            self.image_owner = os.getenv("HB_IMAGE_OWNER", getuser())

        if not self.image_tag:
            self.image_tag = current_commit_sha()

        if not self.nfs_opts:
            self.nfs_opts = os.getenv(
                "HB_NFS_OPTS",
                f"addr=nas.{self.domain},vers=4.1,rw",
            )

    def to_dict(self) -> dict[str, str]:
        assert self.domain
        assert self.name
        assert self.image_registry
        assert self.image_owner
        assert self.image_tag
        assert self.nfs_opts

        stack_env: dict[str, str] = {
            "COMPOSE_PROJECT_NAME": self.name,
            "HB_DOMAIN": self.domain,
            "HB_IMAGE_REGISTRY": self.image_registry,
            "HB_IMAGE_OWNER": self.image_owner,
            "HB_IMAGE_TAG": self.image_tag,
            "HB_NFS_OPTS": self.nfs_opts,
            "NOW": str(int(time())),
        }

        env_update_file: str = os.path.join("secrets", self.domain, ".env")
        if os.path.exists(env_update_file):
            stack_env.update(myke.read.dotfile(env_update_file))

        return stack_env


@lru_cache()
def _get_stack_env(
    name: str | None = None,
    domain: str | None = None,
    image_registry: str | None = None,
    image_owner: str | None = None,
    image_tag: str | None = None,
    nfs_opts: str | None = None,
) -> _StackEnv:
    return _StackEnv(
        name=name,
        domain=domain,
        image_registry=image_registry,
        image_owner=image_owner,
        image_tag=image_tag,
        nfs_opts=nfs_opts,
    )


def stack_config(
    path: str | None = None,
    files: list[str] | None = None,
    domain: str | None = None,
    image_registry: str | None = None,
    image_owner: str | None = None,
    nfs_opts: str | None = None,
    _echo: bool = True,
) -> str:
    if not path:
        path = os.getcwd()

    if not files:
        files = ["docker-compose.yml"]

    with set_cwd(path):
        env_update: dict[str, str] = _get_stack_env(
            name=path,
            domain=domain,
            image_registry=image_registry,
            image_owner=image_owner,
            nfs_opts=nfs_opts,
        ).to_dict()

        stdout, _stderr, _returncode = myke.run(
            (
                "(echo 'version: \"3.9\"' && DOCKER_CONTEXT=default docker compose"
                f" -f {' -f '.join(files)} config) | "
                r' sed "s/^\([[:space:]]\+name:'
                r' [a-zA-Z0-9\_\-]\+\_\)\([[:digit:]]\+\)$/\1$NOW/g"'
            ),
            env_update=env_update,
            capture_output=True,
            echo=_echo,
        )
        assert stdout

        return stdout


def stack_up(
    paths: list[str] | None = yapx.arg(None, pos=True),
    files: list[str] | None = yapx.arg(None),
    domain: str | None = yapx.arg(None, env="HB_DOMAIN"),
    image_owner: str
    | None = yapx.arg(None, flags=["-o", "--owner"], env="HB_IMAGE_OWNER"),
    image_registry: str
    | None = yapx.arg(None, flags=["-r", "--registry"], env="HB_IMAGE_REGISTRY"),
    nfs_opts: str | None = yapx.arg(None, env="HB_NFS_OPTS"),
    no_wait: bool = yapx.arg(False, flags=["-W", "--no-wait"]),
    build: bool = False,
    build_args: dict[str, str] | None = yapx.arg(None, flags=["-a", "--build-args"]),
) -> None:
    if not paths:
        paths = [os.getcwd()]

    for p in paths:
        stack_name: str = _get_stack_name(p, domain=domain)

        if build:
            with set_cwd(p):
                stack_build(
                    image_tags=None,
                    tag_version=None,
                    image_owner=image_owner,
                    target_registry=image_registry,
                    build_args=build_args,
                    force_remote_build=False,
                    no_import=False,
                )

        myke.sh(
            (
                "docker stack deploy --prune --with-registry-auth --resolve-image"
                " always --compose-file - $COMPOSE_PROJECT_NAME"
            ),
            env_update={"COMPOSE_PROJECT_NAME": stack_name},
            input=stack_config(
                p,
                files=files,
                domain=domain,
                image_registry=image_registry,
                image_owner=image_owner,
                nfs_opts=nfs_opts,
                _echo=True,
            ).encode(),
        )

    if not no_wait:
        stack_wait(paths)


def stack_down(
    names: list[str] | None = yapx.arg(None, pos=True),
    no_wait: bool = yapx.arg(False, flags=["-W", "--no-wait"]),
) -> None:
    if not names:
        names = [os.getcwd()]

    names = [_get_stack_name(x) for x in names]

    myke.run(["docker", "stack", "rm", *names])

    if not no_wait:
        stack_wait(names, suppress_errors=True)


def stack_ls(
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
):
    results = _exec_docker_json_query(
        ["stack", "ls"],
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def stack_ps(
    names: list[str] | None = yapx.arg(None, pos=True),
    all_states: bool = False,
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    verbose: bool = False,
    _echo: bool = True,
):
    if not names:
        names = [os.getcwd()]

    names = [_get_stack_name(x) for x in names]

    return service_ps(
        names=None,
        stacks=names,
        all_states=all_states,
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
        no_table=no_table,
        pretty=pretty,
        raw=raw,
        verbose=verbose,
        _echo=_echo,
    )


def stack_build(
    image_tags: list[str] | None = yapx.arg(None, flags=["-t", "--tag", "--tags"]),
    tag_version: str | None = yapx.arg(None, flags=["--tag-version"]),
    image_owner: str
    | None = yapx.arg(None, flags=["-o", "--owner"], env="HB_IMAGE_OWNER"),
    target_registry: str
    | None = yapx.arg(None, flags=["-r", "--registry"], env="HB_IMAGE_REGISTRY"),
    build_args: dict[str, str] | None = yapx.arg(None, flags=["-a", "--build-args"]),
    force_remote_build: bool = yapx.arg(False, flags=["--force-remote-build"]),
    no_import: bool = yapx.arg(False, flags=["-I", "--no-import"]),
):
    if not target_registry:
        target_registry = _get_default_registry()

    if not no_import:
        image_import_file: str = "images.txt"
        if os.path.exists(image_import_file):
            registry_import(
                import_file=image_import_file,
                image_names=None,
                new_registry=target_registry,
                new_owner=None,
                all_tags=False,
                skopeo_image_registry=None,
            )

    tag_list: list[str] = [current_commit_sha()]

    if tag_version:
        tag_list.append(tag_version)
        try:
            version = packaging.version.parse(tag_version)
            if not any(
                [version.is_devrelease, version.is_prerelease, version.is_postrelease],
            ):
                for i in 3, 2, 1:
                    if i <= len(version.release):
                        v: str = ".".join(str(x) for x in version.release[:i])
                        if v and v not in tag_list:
                            tag_list.append(v)
        except packaging.version.InvalidVersion:
            pass

    if image_tags:
        tag_list.extend(image_tags)

    if not image_owner:
        image_owner = getuser()

    destinations: str = "--destination " + " --destination ".join(
        f"{target_registry}/{image_owner}/{_get_stack_name()}:{x}" for x in tag_list
    )

    if not build_args:
        build_args = {}

    build_args["HB_IMAGE_REGISTRY"] = target_registry

    build_args_str: str = " ".join(
        f"--build-arg {k}={v}" for k, v in build_args.items()
    )

    myke.run(
        _kaniko_cli(force_remote_build=force_remote_build)
        + f"--build-arg HB_IMAGE_REGISTRY='{target_registry}'"
        f" {build_args_str} {destinations}",
    )


def stack_create(
    directory: str,
    minimal: bool = False,
    pre_release: bool = yapx.arg(False, flags=["--pre", "--pre-release"]),
    answers: dict[str, str] | None = None,
    template_repo: str = yapx.arg(
        "https://codeberg.org/Fresh2dev/copier-f2dv-project.git",
        flags=["-r", "--repo"],
    ),
    dry_run: bool = False,
    force: bool = False,
):
    copier_args: list[str] = ["--overwrite"]

    if dry_run:
        copier_args.append("--pretend")

    if pre_release:
        copier_args.extend(["--vcs-ref", "dev"])

    if not force and os.path.exists(directory) and os.listdir(directory):
        raise FileExistsError("Refusing to target non-empty directory: " + directory)

    copier_answers: dict[str, str] = {
        "is_python": "false",
        "is_minimal": "true" if minimal else "false",
    }

    if answers:
        copier_answers.update(answers)

    copier_args.extend(
        x for k, v in copier_answers.items() for x in ["--data", f"{k}={v}"]
    )

    myke.run(
        ["copier"]
        + copier_args
        + ["copy"]
        + ["--no-cleanup", template_repo, directory],
        env_update={"HOME": "/dev/null"},
    )


def stack_update(
    answers: dict[str, str] | None = None,
    pre_release: bool = yapx.arg(False, flags=["-p", "--pre", "--pre-release"]),
    clobber: bool = False,
    dry_run: bool = False,
):
    copier_answers = myke.read.yaml(".copier-answers.yml")
    template_repo: str = copier_answers.pop("_src_path")
    copier_answers = {k: v for k, v in copier_answers.items() if not k.startswith("_")}

    if answers:
        copier_answers.update(answers)

    if clobber:
        stack_create(
            directory=os.getcwd(),
            pre_release=pre_release,
            answers=copier_answers,
            template_repo=template_repo,
            dry_run=dry_run,
            force=True,
        )
    else:
        copier_args: list[str] = ["--overwrite"]

        if dry_run:
            copier_args.append("--pretend")

        if pre_release:
            copier_args.extend(["--vcs-ref", "dev"])

        copier_args.extend(
            x for k, v in copier_answers.items() for x in ["--data", f"{k}={v}"]
        )

        myke.run(
            ["copier"] + copier_args + ["update"],
            env_update={"HOME": "/dev/null"},
        )

    if not dry_run:
        myke.run("find . -name '*.rej' -delete;")

        print("Pulled latest modifications. Running pre-commit...\n")
        myke.run("git pre-commit --all-files", echo=False, check=False)
        print("Update complete. Review changes before committing.\n")


def stack_secrets(
    vault_server: str = yapx.arg(env=["HB_VAULT_ADDR", "VAULT_ADDR"]),
    vault_api_token: str = yapx.arg(env=["HB_VAULT_TOKEN", "VAULT_TOKEN"]),
    get_secrets: bool = yapx.arg(False, exclusive=True, flags=["--get"]),
    put_secrets: bool = yapx.arg(False, exclusive=True, flags=["--put"]),
    put_all_secrets: bool = yapx.arg(False, exclusive=True, flags=["--put-all"]),
    domains: list[str] | None = yapx.arg(None, env="HB_DOMAIN"),
    secrets_dir: str = "secrets",
    vault_root_path: str | None = "secret/hostbutter",
):
    if not os.path.exists(secrets_dir):
        raise FileNotFoundError(secrets_dir)

    if put_all_secrets:
        put_secrets = True
        if not domains:
            domains = os.listdir(secrets_dir)
    elif not domains:
        domains = [os.environ["DOCKER_CONTEXT"]]

    with set_env(VAULT_ADDR=vault_server, VAULT_TOKEN=vault_api_token):
        for d in domains:
            domain_secrets = os.path.join(secrets_dir, d)

            vault_path = f"projects/{d}/{current_project_name()}"

            if get_secrets:
                ci_secrets(
                    vault_server=vault_server,
                    vault_api_token=vault_api_token,
                    get_secrets=True,
                    put_secrets=None,
                    patch_secrets=None,
                    scope=None,
                    vault_path=vault_path,
                    vault_root_path=vault_root_path,
                )
            elif put_secrets:
                secret_files: dict[str, str] = {
                    x: "@" + os.path.join(domain_secrets, x)
                    for x in os.listdir(domain_secrets)
                    if os.path.isfile(x)
                }

                if secret_files:
                    ci_secrets(
                        vault_server=vault_server,
                        vault_api_token=vault_api_token,
                        get_secrets=False,
                        put_secrets=secret_files,
                        patch_secrets=None,
                        scope=None,
                        vault_path=vault_path,
                        vault_root_path=vault_root_path,
                    )
            else:
                print("Specify '--get' or '--put'")
                break



from __future__ import annotations

import os
import sys

import myke
import yaml
import yapx

from ..__version__ import __version__
from ..exceptions import NoKnownCiServerError, NoKnownDnsServerError
from ..utils import current_project_name, set_env
from .context import context_current
from .registry import _get_default_registry

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


def _get_default_ci_server(context: str | None = None) -> str:
    reg: str | None = os.getenv("DRONE_SERVER")

    if not reg:
        if not context:
            context = context_current(from_env=True, _echo=False)

        if "." not in context:
            raise NoKnownCiServerError(
                "Unable to infer CI server URL within current context: " + context,
            )

        reg = "ci." + context

    return reg


def ci_url(
    # ci_api_token: str = yapx.arg(env=['HB_CI_TOKEN', 'DRONE_TOKEN']),
    ci_server: str
    | None = yapx.arg(None, env=["HB_CI_SERVER", "DRONE_SERVER"]),
):
    with set_env(DRONE_SERVER=ci_server):
        if not ci_server:
            ci_server = _get_default_ci_server()

        if not ci_server.startswith("https://"):
            ci_server = "https://" + ci_server

        print(f"{ci_server.rstrip('/')}/{current_project_name()}")


def ci_template(
    image_registry: str
    | None = yapx.arg(None, flags=["-r", "--registry"], env="HB_IMAGE_REGISTRY"),
    dns_servers: list[str] | None = yapx.arg(None, flags=["--dns", "--dns-servers"]),
):
    if not dns_servers:
        with open(os.environ["ANSIBLE_INVENTORY"], encoding="utf-8") as f:
            dns_servers = (
                yaml.safe_load(f)
                .get("all", {})
                .get("vars", {})
                .get("dnsmasq_dns_servers")
            )

        if not dns_servers:
            raise NoKnownDnsServerError()

    if not image_registry:
        image_registry = _get_default_registry()

    myke.echo.text(
        myke.read.text(
            os.path.join(
                os.path.dirname(os.path.dirname(__file__)),
                "resources",
                "hostbutter.jsonnet",
            ),
        )
        .replace("%HB_IMAGE_REGISTRY%", image_registry, 1)
        .replace("%HB_CI_DNS%", "', '".join(x.strip("'\",") for x in dns_servers), 1)
        .replace("%HB_VERSION%", __version__, 1),
    )


def ci_secrets(
    vault_server: str = yapx.arg(env=["HB_VAULT_ADDR", "VAULT_ADDR"]),
    vault_api_token: str = yapx.arg(env=["HB_VAULT_TOKEN", "VAULT_TOKEN"]),
    get_secrets: bool = yapx.arg(False, exclusive=True, flags=["--get"]),
    put_secrets: dict[str, str]
    | None = yapx.arg(None, exclusive=True, flags=["--put"]),
    patch_secrets: dict[str, str]
    | None = yapx.arg(None, exclusive=True, flags=["--patch"]),
    scope: Literal["global", "cluster"] | None = None,
    vault_path: str | None = None,
    vault_root_path: str | None = "secret/hostbutter",
):
    vault_root_path.rstrip("/")

    if scope == "global":
        vault_root_path = f"{vault_root_path}/global"
    elif scope == "cluster":
        vault_root_path = f"{vault_root_path}/clusters/{os.environ['DOCKER_CONTEXT']}"

    if not vault_path:
        vault_path = vault_root_path
    else:
        vault_path = f"{vault_root_path}/{vault_path.lstrip('/')}"

    if patch_secrets:
        put_secrets = patch_secrets

    with set_env(VAULT_ADDR=vault_server, VAULT_TOKEN=vault_api_token):
        if get_secrets:
            myke.run(["vault", "kv", "get", "-format=yaml", vault_path], check=False)
        elif put_secrets:
            myke.run(
                [
                    "vault",
                    "kv",
                    "patch" if patch_secrets else "put",
                    "--non-interactive",
                    vault_path,
                ]
                + [f"{k}={v}" for k, v in put_secrets.items()],
                check=False,
            )
        else:
            print("Specify '--get' or '--put'")



from typing import List, Pattern

import mocx
from _pytest.capture import CaptureFixture, CaptureResult

from hostbutter.__main__ import main
from hostbutter.__main__ import sys as target_sys


def test_main_version(capsys: CaptureFixture, version_regex: Pattern):
    # 1. ARRANGE
    args: List[str] = ["version"]

    # 2. ACT
    with mocx.patch.object(target_sys, "argv", [""] + args):
        main()

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out
    assert version_regex.search(captured.out), "invalid version"


# def test_tmp(capsys: CaptureFixture):
#     # 1. ARRANGE
#     # args: List[str] = ["image-tags", "docker.io/python:latest", "--filter", "3.10.[0-9]*-*"]
#     args: List[str] = ["-c", "default", "registry-ls"]
#     # args: List[str] = ["registry-images", "registry.lokalhost.net"]
#     # args: List[str] = ["image-inspect", "docker.io/python:3.10.8"]

#     # 2. ACT
#     with mocx.patch.object(target_sys, "argv", [""] + args):
#         main()

#     # 3. ASSERT
#     captured: CaptureResult = capsys.readouterr()
#     assert not captured.err
#     assert captured.out



from __future__ import annotations

import os
from typing import Any

import myke
import yapx

from ..query import _exec_docker_json_query, _print_results


def context_ls(
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
):
    jq.insert(
        0,
        (
            # r'.[] | select(.DockerEndpoint | startswith("ssh://"))'
            r".[] | select(.DockerEndpoint)"
            r" | {Name: .Name, DockerEndpoint: .DockerEndpoint}"
        ),
    )

    results: Any = _exec_docker_json_query(
        cmd=["context", "ls"],
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def context_current(
    from_env: bool = yapx.arg(False, flags=["-e", "--from-env"]),
    _echo: bool = True,
) -> str:
    ctx: str | None = os.getenv("DOCKER_CONTEXT") if from_env else None

    if not ctx:
        default_context: str = "default"

        try:
            ctx = myke.read.json(os.environ["DOCKER_CONFIG_FILE"]).get(
                "currentContext",
                default_context,
            )
        except FileNotFoundError:
            ctx = default_context

        assert ctx

    if _echo:
        print(ctx)

    return ctx


def context_default(name: str = yapx.arg(pos=True)):
    myke.run(
        ["docker", "context", "use", name],
        env_update={"DOCKER_CONTEXT": "default"},
    )



from __future__ import annotations

from typing import Any

import myke
import yapx

from ..constants import SWARM_GROUP_NAME
from ..query import (
    _exec_ansible_docker_json_query,
    _exec_docker_json_query,
    _exec_json_query,
    _print_results,
)
from .cluster import cluster_exec, cluster_ping, cluster_playbook
from .context import context_current, context_ls


def swarm_ssh(
    name: str | None = yapx.arg(None, pos=True),
    dry_run: bool = False,
    _echo: bool = True,
) -> str:
    ctx: str = context_current(from_env=True, _echo=False)
    if not name:
        name = ctx
    elif "." not in name and "." in ctx:
        name += "." + ctx

    matches: list[str] = context_ls(
        jq=[f'.[] | select (.Name == "{name}")'],
        no_table=True,
        _echo=False,
        extract="DockerEndpoint",
    )

    if not matches:
        print("Not a recognized context. Select one of:")
        context_ls(jq=[], no_table=False)
        raise ValueError(name)

    ssh_target: str = matches[0]

    if _echo:
        print(ssh_target)

    if not dry_run:
        myke.run(["ssh", ssh_target])

    return ssh_target


def swarm_exec(command: str):
    cluster_exec(command=command, limit_to=SWARM_GROUP_NAME)


def swarm_query(
    command: list[str] = yapx.arg(pos=True),
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
) -> list[dict[str, Any] | Any]:
    results: Any = _exec_docker_json_query(
        cmd=command,
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def swarm_nodes(
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
):
    # ID: .ID
    # , vCPUs: .Description.Resources.NanoCPUs
    # , MemGB: .Description.Resources.MemoryBytes
    # , MgrAddr: .ManagerStatus.Addr
    # , Labels: .Spec.Labels
    jq.insert(
        0,
        r""".[] | {
    HostName: .Description.Hostname
    , Engine: .Description.Engine.EngineVersion
    , Arch: .Description.Platform.Architecture
    , Role: .Spec.Role
    , Availability: .Spec.Availability
    , State: .Status.State
    , IP: .Status.Addr
    , MgrReachability: .ManagerStatus.Reachability
    , MgrLeader: .ManagerStatus.Leader
}""",
    )

    results = _exec_json_query(
        cmd="docker node ls -q | xargs -n1 docker node inspect --format "
        + r"'{{ json . }}'",
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def swarm_labels(
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
):
    # ID: .ID
    # , vCPUs: .Description.Resources.NanoCPUs
    # , MemGB: .Description.Resources.MemoryBytes
    # , MgrAddr: .ManagerStatus.Addr
    jq.insert(
        0,
        (
            r".[] | [.Spec.Labels | to_entries"
            r" | map({Label: .key, Value: .value})][][] + {Node: .Description.Hostname}"
        ),
    )

    results = _exec_json_query(
        cmd="docker node ls -q | xargs -n1 docker node inspect --format "
        + r"'{{ json . }}'",
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def swarm_networks(
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
    _echo: bool = True,
):
    jq.insert(
        0,
        (
            r".[] | {Name: .Name, Subnet: .IPAM.Config[]?.Subnet, Containers:"
            r" .Containers | length, Peers: .Peers | length, Scope: .Scope, Driver:"
            r" .Driver, Internal: .Internal, Ingress: .Ingress, Attachable:"
            r" .Attachable}"
        ),
    )

    results = _exec_json_query(
        cmd=(
            r"docker network ls --format '{{ .Name }}'"
            r" | xargs -n1 docker network inspect --format '{{ json . }}'"
        ),
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if _echo:
        if extract:
            raw = True
        _print_results(results, no_table=no_table, pretty=pretty, raw=raw)

    return results


def swarm_ping():
    cluster_ping(limit_to=SWARM_GROUP_NAME)


def swarm_prune():
    swarm_exec(
        command=(
            'docker system prune --all --volumes --force && (for x in "volume"; do'
            " docker $x rm $(docker $x ls -q) 2>/dev/null || true; done)"
        ),
    )


def swarm_stats(
    verbose: bool = False,
    jq: list[str] = yapx.arg(lambda: [], flags=["-q", "--jq"]),
    select: list[str] | None = None,
    drop: list[str] | None = None,
    extract: str | None = None,
    sort: list[str] | None = None,
    sort_reverse: bool = False,
    unique: bool = False,
    no_table: bool = yapx.arg(False, flags=["-T", "--no-table"]),
    pretty: bool = False,
    raw: bool = False,
):
    jq_base: str = (
        r".[] | {"
        r" HostName: .__hostname__"
        r", Task: .Name"
        r", Container"
        r", CPUPerc: (.CPUPerc[:-1] | tonumber)"
        r", MemPerc: (.MemPerc[:-1] | tonumber)"
        r", MemUsage"
        r", ID, BlockIO, NetIO, PIDs"
        r"}"
    )

    if not verbose:
        jq_base += r" | del(.ID, .BlockIO, .NetIO, .PIDs)"

    jq = ([jq_base] + jq) if jq else [jq_base]

    results = _exec_ansible_docker_json_query(
        ["stats", "--no-stream"],
        limit_to=SWARM_GROUP_NAME,
        jq=jq,
        select=select,
        drop=drop,
        extract=extract,
        sort=sort,
        sort_reverse=sort_reverse,
        unique=unique,
    )

    if extract:
        raw = True
    _print_results(results, no_table=no_table, pretty=pretty, raw=raw)


def swarm_reboot():
    cluster_playbook("fresh2dev.hostbutter.reboot_swarm", limit_to=SWARM_GROUP_NAME)



class NoTagsFoundError(Exception):
    ...


class TooManyTagsError(Exception):
    ...


class NoKnownRegistryError(Exception):
    ...


class NoKnownCiServerError(Exception):
    ...


class NoKnownDnsServerError(Exception):
    ...
